/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

// aphot.c: aperture photometry routines
// $Revision: 1.2 $
// $Date: 2004/02/12 01:00:15 $

#include <stdio.h>
#include <stdlib.h>
#include <sys/io.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>
#include <string.h>

#include "ccd.h"
//#include "x11ops.h"


// compute ring statistics between r1 and r2; ring has the origin at x, y;
// values lower than min or larger than max are skipped
// quads contains a bitfield enabling the measurement of quadrants; bit 0 is Q0 (dx>=0,dy>=0), 
// bit 1 is Q1 (dx<0, dy>=0) etc. 
int ring_stats(struct ccd_frame *fr, double x, double y, 
	       double r1, double r2, int quads, struct rstats *rs,
	       double min, double max)
{
	int ix, iy;
	int xs, ys;
	int xe, ye;
	int xc, yc;
	int w;
	int nring = 0, nskipped = 0;
	float *dp;
	float v;
	double sum, sumsq, r;
//	d3_printf("x:%f y:%f r1:%f r2:%f\n", x, y, r1, r2);

// round the coords off to make sure we use the same number of pixels for the
// flux measurement

	xc = (int)floor(x+0.5);
	yc = (int)floor(y+0.5); 

	xs = (int)(xc-r2); 
	xe = (int)(xc+r2+1); 
	ys = (int)(yc-r2); 
	ye = (int)(yc+r2+1); 
	w = fr->w;

	if (xs < 0)
		xs = 0;
	if (ys < 0)
		ys = 0;
	if (xe >= w)
		xe = w - 1;
	if (ye >= fr->h)
		ye = fr->h - 1;

//	d3_printf("xc:%d yc:%d r1:%f r2:%f\n", xc, yc, r1, r2);

//	d3_printf("xs:%d xe:%d ys:%d ye:%d w:%d\n", xs, xe, ys, ye, w);

//	d3_printf("enter ringstats\n");

	for (ix = 0; ix < H_SIZE; ix++)
		(rs->h[ix]) = 0;

//	d3_printf("init histogram\n");

// walk the ring and collect statistics
	sum=0.0;
	sumsq = 0.0;
	rs->min = HUGE;
	rs->max = -HUGE;
	dp = ((float *)(fr->dat)) + xs + ys * w;
	for (iy = ys; iy < ye; iy++) {
		for (ix = xs; ix < xe; ix++) {
			v = *dp;
			dp ++;
			if (((r = sqr(ix - xc + 0.0) + sqr(iy - yc + 0.0)) 
			     < sqr(r1) || r >= sqr(r2)))
				continue;
			if (((quads & QUAD1) == 0) && ix-xc >= 0 && iy-yc >= 0)
				continue;
			if (((quads & QUAD2) == 0) && ix-xc < 0 && iy-yc >= 0)
				continue;
			if (((quads & QUAD3) == 0) && ix-xc < 0 && iy-yc > 0)
				continue;
			if (((quads & QUAD4) == 0) && ix-xc >= 0 && iy-yc < 0)
				continue;
			nring ++;
			if (v < H_START || v > H_START+H_SIZE) {
				nskipped ++;
				continue;
			}
			if (v < min || v > max) {
				nskipped ++;
				continue;
			}
			rs->h[(int)floor(v) - H_START] ++;
			sum += v;
			sumsq += v*v;
			if (v < rs->min) 
				rs->min = v;
			if (v > rs->max) 
				rs->max = v;
		}
		dp += w - xe + xs;
	}
// fill the result structure with the stats
	rs->all = nring;
	rs->used = nring - nskipped;
	rs->sum = sum;
	rs->sumsq = sumsq;
	rs->avg = sum / rs->used;
	rs->sigma = sqrt(sumsq / rs->used - sqr(sum / rs->used));
// walk the histogram and compute median
	iy = 0;
	for (ix = 0; ix < H_SIZE; ix++) {
		iy += rs->h[ix];
		if (iy >= rs->used / 2)
			break;
	}
//	d3_printf("ix = %d iy = %d\n", ix, iy);
	rs->median = ix + H_START;
//	d3_printf("ringstats avg:%.2f sum:%.2f sigma%.2f, min%.2f max%.2f dx%.2f dy%.2f", 
//		  rs->avg, rs->sum, rs->sigma, rs->min, rs->max, rs->dx, rs->dy); 
	return 0;
}

// check the photometry params and fill up with reasonable values
// return 0 if all values checked ok, 1 if changes were made
/*
static int check_ap_params(struct ap_params *p)
{
	int ret = 0;
	if (p->sigmas <= 0.0) {
		p->sigmas = 1.0;
		ret = 1;
	}
	if (p->max_iter <= 0) {
		p->max_iter = 8;
		ret = 1;
	}
	if (p->r1 < 1.0) {
		p->r1 = 1.0;
		ret = 1;
	}
	if (p->r2 < p->r1 + 1.0) {
		p->r2 = p->r1 + 1.0;
		ret = 1;
	}
	if (p->r3 < p->r2 + 1.0) {
		p->r3 = p->r2 + 1.0;
		ret = 1;
	}
	if (p->sat_limit <= 0) {
		p->sat_limit = 50000;
		ret = 1;
	}
	return ret;
}
*/

// get the sky value and error; we assume that the star position 
// has been checked against the edges
static int ap_get_sky(struct ccd_frame *fr, struct star *s, 
	       struct ap_params *p, struct bad_pix_map *bp)
{
	struct rstats rs;
	int ret, i;
	double phn, omed, med, min, max, flnsq;

// get a first round of ring stats
	ret = ring_stats(fr, s->x, s->y, p->r2, p->r3, p->quads, &rs, -HUGE, HUGE);
	if (ret) {
		return ret;
	}

	switch(p->sky_method) {
	case APMET_AVG:
		s->aph.sky = rs.avg;
		break;
	case APMET_MEDIAN:
		s->aph.sky = rs.median;
		break;
	case APMET_MMEDIAN:
		i=0;
		omed = rs.median;
//		do {
			med = rs.median;
			min = med - p->sigmas * rs.sigma;
			max = med + p->sigmas * rs.sigma;
			ret = ring_stats(fr, s->x, s->y, p->r2, p->r3, p->quads, &rs, min, max);
			if (ret) {
				err_printf("ap_get_sky: error in ring_stats\n");
				return ret;
			}
//		} while(i < p->max_iter && rs.median != med);
//		if (i == p->max_iter) {
//			d3_printf("mmedian did not converge after %d iterations %.1f, %.1f\n",
//				  omed, rs.median); 
//			s->flags |= AP_DID_NOT_CONVERGE;
//		}
//		d3_printf("sky omed: %.1f, med:%.1f, mm: %.1f\n", omed, med, rs.avg);
		s->aph.sky = rs.avg;
		break;
	default:
		err_printf("ap_get_sky: unknown method %d\n", p->sky_method);
		free(rs.h);
		return ERR_FATAL;
	}

// copy stats to the star structure
	s->aph.sky_avg = rs.avg;
	s->aph.sky_min = rs.min;
	s->aph.sky_max = rs.max;
	s->aph.sky_med = rs.median;
	s->aph.sky_sigma = rs.sigma;
	s->aph.sky_all = rs.all;
	s->aph.sky_used = rs.used;

// compute background error
	phn = sqrt(fabs(s->aph.sky - fr->exp.bias) * fr->exp.scale) / fr->exp.scale; // sky+dark photon shot noise / pixel
	flnsq = sqr(fr->exp.flat_noise) * rs.sumsq;
	s->aph.sky_err = sqrt((sqr(fr->exp.rdnoise) + sqr(phn) + flnsq) / s->aph.sky_used);
	return 0;
}

// get the star flux
static int ap_get_star(struct ccd_frame *fr, struct star *s, 
	       struct ap_params *p, struct bad_pix_map *bp)
{
	struct rstats rs;
	int ret;
	double phn, flnsq;

	ret = ring_stats(fr, s->x, s->y, 0, p->r1, QUAD1|QUAD2|QUAD3|QUAD4, &rs, -HUGE, HUGE);
	if (ret) {
		return ret;
	}

// copy stats to star structure
	s->aph.tflux = rs.sum;
	s->aph.star_min = rs.min;
	s->aph.star_max = rs.max;
	s->aph.star_all = rs.used;

// see if any flags need to be set
	if (rs.used != rs.all)
		s->aph.flags |= AP_STAR_SKIP;
	if (rs.max > p->sat_limit)
		s->aph.flags |= AP_BURNOUT;
	else if (rs.max > p->sat_limit / 2)
		s->aph.flags |= AP_BRIGHT;

	phn = sqrt(fabs(s->aph.tflux - fr->exp.bias * s->aph.star_all) 
		   * fr->exp.scale) / fr->exp.scale; // total photon shot noise
	flnsq = sqr(fr->exp.flat_noise) * (rs.sumsq);
	s->aph.flux_err = sqrt((sqr(fr->exp.rdnoise) * s->aph.star_all + sqr(phn) + flnsq));
	s->aph.pshot_noise = phn;
	s->aph.rd_noise = fr->exp.rdnoise * sqrt(s->aph.star_all);
//	d3_printf("pshotn %.4f fluxerr %.4f flat %.4f\n", phn, s->flux_err, sqrt(flnsq));
	return 0;
}


// do aperture photometry for star s; star image coordinates are taken from s. The function
// fills up the results in s. If an error (!= 0) is returned, the values in s are undefined.
// the bad pix map is used for flagging possible data errors due to the presence of bad pixels
int aperture_photometry(struct ccd_frame *fr, struct star *s, 
			struct ap_params *p, struct bad_pix_map *bp)
{
	double rm;
	int ret;
	double skn, rdn;

	rm = ceil(p->r3) + 1;

//	d3_printf("rm = %f\n", rm);
	if ((s->x - rm) <= 0 || (s->x + rm) >= fr->w ||
	    (s->y - rm) <= 0 || (s->y + rm) >= fr->h) {
		err_printf("aperture_photometry: star is too close to the frame edge\n");
		return ERR_FATAL;
	}

// get the sky
	ret = ap_get_sky(fr, s, p, bp);
	if (ret)
		return ret;

// now get the data for the star
	ret = ap_get_star(fr, s, p, bp);
	if (ret)
		return ret;

// compute star flux and error
	s->aph.star = s->aph.tflux - s->aph.star_all * s->aph.sky;
	s->aph.star_err = sqrt(sqr(s->aph.flux_err) + sqr(s->aph.star_all * s->aph.sky_err));

	if (s->aph.star < s->aph.star_err)
		s->aph.flags |= AP_FAINT;
// clip to protect the logarithms
	if (s->aph.star < MIN_STAR)
		s->aph.star = MIN_STAR;
	if (s->aph.star_err < MIN_STAR)
		s->aph.star_err = MIN_STAR;

	s->aph.absmag = -2.511886*log10(s->aph.star);

// compute limiting magnitude as the magnitude at which the SNR = 1
	skn = sqr(s->aph.star_all * s->aph.sky_err);
	rdn = s->aph.star_all * sqr(fr->exp.rdnoise);
//	d3_printf("star error before log: %.3f\n", s->star_err / s->star);
// temporary magerr that doesn't include scintillation
	s->aph.magerr = 2.511886 * (log10(s->aph.star + s->aph.star_err) - log10(s->aph.star));
	s->aph.flags |= AP_MEASURED;
	return 0;
}

// compute absolute magnitude from flux
double flux_to_absmag(double flux)
{
	return -2.511886*log10(flux);
}

/*
static void print_star_results(struct ph_star *s)
{
	info_printf("imag:%8.3f err:%8.3f\n", s->absmag, s->magerr);
	info_printf("Sky pixels:%f used:%f value:%.2f error:%.2f\n", 
		    s->sky_all, s->sky_used, s->sky, s->sky_err); 
	info_printf("Star pixels:%f tflux:%.2f flux:%.2f err:%.2f mag:%.2f\n",
		    s->star_all, s->tflux, s->star, s->star_err, s->absmag);
	info_printf("Star min:%.2f max:%.2f flags:%x\n",
		    s->star_min, s->star_max, s->flags);
}
*/

// get scintillation noise parameters
void get_scint_pars(struct ccd_frame *fr, struct vs_recipy *vs, double *t, double *d, double *am)
{

	FITS_row * cp;
	float r;

// determine exposure time

	if (fits_get_double(fr, "EXPTIME", t) <= 0 ) {
		err_printf("bad exposure time, assuming 1 sec\n");
		*t = 1.0;
	}
// telescope aperture
	cp = fits_keyword_lookup(fr, "APERT");
	if (vs->aperture > 1.0 && vs->aperture < 1000) { // look in recipy
		*d = vs->aperture;
	} else if (cp != NULL && (sscanf((char *)cp, "%*s = %f", &r) )) {
		*d = r;
	} else {
		err_printf("bad telescope aperture, assuming 30 cm\n");
		*d = 30.0;
	}
// airmass
	cp = fits_keyword_lookup(fr, "OBJCTALT");
	if (cp != NULL && (sscanf(((char *)cp), "%*s= %f", &r) )) {
		*am = airmass(r);
		return;
	}
	cp = fits_keyword_lookup(fr, "AIRMASS");
	if (cp != NULL && (sscanf(((char *)cp), "%*s = %f", &r) )) {
		*am = r;
		return;
	}
	if (vs->airmass >= 1.0 && vs->airmass < 50.0) { // we get it from here
		*am = vs->airmass;
		return;
	}
	err_printf("could not determine airmass, assuming 1.5\n");
	*am = 1.5;
}

double scintillation(double t, double d, double am)
{
	double sn;
//	d3_printf("t %.1f d %.1f am %.1f", t, d, am);
	sn = log(0.09) - 0.66666 * log(d) + 1.75 * log(am) - 0.5 * log(2 * t);
	return exp(sn);
}

// compute 1-sigma scintillation noise
double scint_noise(struct ccd_frame *fr, struct vs_recipy *vs)
{
	double t, d, am;

	get_scint_pars(fr, vs, &t, &d, &am);
	return vs->scint_factor * scintillation(t, d, am);
}


// max shift of a star from the std position 
#define MAX_SHIFT 3
// measure the stars in a recipy; return 0 if no error occured
int measure_stars(struct ccd_frame *fr, struct vs_recipy *vs)
{
	int i;
	struct star *s;
	struct star st;
	int ret;
	double snsq;

	ret = 0;

	snsq = scint_noise(fr, vs);
	snsq = sqr(snsq);

	for(i = 0; i < vs->cnt; i++) {
		s = &(vs->s[i]);
		if (vs->usewcs) {
			xypix(s->ra, s->dec, fr->fim.xref, fr->fim.yref, 

			      fr->fim.xrefpix, fr->fim.yrefpix, fr->fim.xinc, 
			      fr->fim.yinc, fr->fim.rot,
			      "-TAN", &s->x, &s->y);
			info_printf("Star near %.1f %.1f ", s->x, s->y);
			get_star_near(fr, (int)s->x, (int)s->y, 0, &st);
			info_printf("found at %.1f %.1f ", st.x, st.y) ;
//			st.x = s->x;
//			st.y = s->y;
			if (fabs(s->x - st.x) > MAX_SHIFT || fabs(s->y - st.y) > MAX_SHIFT) {
				info_printf(" ... Cannot locate star, too faint?\n");
			} else {
				s->x = st.x;
				s->y = st.y;
			}
		}
		ret |= aperture_photometry(fr, s, &(vs->p), NULL);
//		print_star_results(s);
//		d3_printf("magerr before scint: %.4f\n", s->magerr);
		s->aph.magerr = sqrt(sqr(s->aph.magerr) + snsq);
//		info_printf("after scint err:%8.3f\n", s->magerr);
	}
	return ret;

}

// compute a photometry solution for the given recipy
void get_ph_solution(struct vs_recipy *vs)
{
	int i;
	double std_cal;
	double nstd;
	double weight;
	double std_cal_err;

	nstd = 0.0;
	std_cal = 0.0;
	std_cal_err = 0.0;

	for (i=0; i<vs->cnt; i++) {
		if (vs->s[i].aph.flags & AP_STD_STAR) {
			if (vs->s[i].aph.magerr > 0.0001) {
				weight = 1 / vs->s[i].aph.magerr;
			} else {
				weight = 1 / 0.0001;
			}
			nstd += weight;
			std_cal += weight * (vs->s[i].aph.stdmag - vs->s[i].aph.absmag);
			std_cal_err += sqr(weight * vs->s[i].aph.magerr);
		}
	}
	if (nstd == 0) {
		err_printf("No standard stars!\n");
		return;
	}
	std_cal = std_cal / nstd;
	std_cal_err = sqrt(std_cal_err) / nstd;
//	d3_printf("std_cal_err : %.4f\n", std_cal_err);
	vs->p.std_err = std_cal_err;
	for (i=0; i<vs->cnt; i++) {
		if ((vs->s[i].aph.flags & AP_STD_STAR) == 0) {
			vs->s[i].aph.stdmag = vs->s[i].aph.absmag + std_cal;
			vs->s[i].aph.magerr = sqrt(sqr(vs->s[i].aph.magerr) + sqr(std_cal_err));
		} else {
			vs->s[i].aph.residual = vs->s[i].aph.absmag 
				+ std_cal - vs->s[i].aph.stdmag;
		}
	}
}


