/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

/* Observation data handling */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include "gcx.h"
#include "camera.h"
#include "catalogs.h"
#include "obsdata.h"
#include "params.h"

#include "sidereal_time.h"

/* extract the catalog name (the part before ":") from the name and return 
 * a pointer to it (a terminating 0 is added). Also update name
 * to point to the object name proper */
char *extract_catname(char *text, char **oname)
{
	char *catn = NULL, *n;
	n = text;
	while (*n != 0) {
		if (!isspace(*n)) {
			catn = n;
			break;
		}
		n++;
	}
	while (*n != 0) {
		if (*n == ':') {
			break;
		}
		n++;
	}
	if (*n == 0) {
		*oname = catn;
		return NULL;
	} else {
		*n = 0;
		n++;
	}
	while (*n != 0) {
		if (!isspace(*n)) {
			break;
		}
		n++;
	}
	*oname = n;
	return catn;
}

/* replace a malloced string with the argument */
void replace_strval(char **str, char *val)
{
	if (*str)
		free(*str);
	*str = val;
}

/* get a catalog object cats. the object is specified by 
 * a "catalog:name" string. If the catalog part is not present, 
 * edb is searched.
 * return cats if successfull, NULL if some error occured.
 */

struct cat_star *get_object_by_name(char *name)
{
	char *catn, *oname;
	struct catalog *cat;
	struct cat_star *cats;
	int ret;

	if (name == NULL)
		return NULL;
	catn = extract_catname(name, &oname);
	d3_printf("catn |%s| name |%s|\n", catn, oname);
	if (catn == NULL)
		catn = "edb";
	cat = open_catalog(catn);
	if (cat == NULL)
		return NULL;
	if (cat->cat_get == NULL)
		return NULL;
	ret = (*(cat->cat_get))(&cats, cat, oname, 1);
	close_catalog(cat);
	if (ret <= 0)
		return NULL;
	return cats;
}


/* get a catalog object into the obs. the object is specified by 
 * a "catalog:name" string. If the catalog part is not present, 
 * edb is searched.
 * return 0 if successfull, -1 if some error occured.
 */
int obs_set_from_object(struct obs_data *obs, char *name)
{
	struct cat_star *cats;
	cats = get_object_by_name(name);
	if (cats == NULL)
		return -1;
	obs->ra = cats->ra;
	obs->dec = cats->dec;
	obs->equinox = cats->equinox;
	replace_strval(&(obs->objname), strdup(name));
	cat_star_release(cats);
	return 0;
}

/* object creation/destruction */
struct obs_data * obs_data_new(void)
{
	struct obs_data *obs;
	obs = calloc(1, sizeof(struct obs_data));
	if (obs == NULL)
		return NULL;
	obs->ref_count = 1;
	obs->rot = 0.0;
	return obs;
}

void obs_data_ref(struct obs_data *obs)
{
	if (obs == NULL)
		return;
	obs->ref_count ++;
}

void obs_data_release(struct obs_data *obs)
{
	if (obs == NULL)
		return;
	if (obs->ref_count < 1)
		err_printf("obs_data_release: obs_data has ref_count of %d\n", obs->ref_count);
	if (obs->ref_count <= 1) {
		if (obs->objname)
			free(obs->objname);
		if (obs->filter)
			free(obs->filter);
		if (obs->comment)
			free(obs->filter);
		free(obs);
	} else {
		obs->ref_count --;
	}
}

/* Add obs and information to the frame */
/* If obs is null, only the global obs data (from PAR) is added */

void ccd_frame_add_obs_info(struct ccd_frame *fr, struct obs_data *obs)
{
	char lb[128];
	char deg[64];

	if (obs != NULL) {
		if (obs->ra != 0.0 || obs->rot != 0.0 || obs->dec != 0.0) {
			fr->fim.xref = obs->ra;
			fr->fim.yref = obs->dec;
			fr->fim.rot = obs->rot;
			fr->fim.wcsset = 1;
		}
		fr->fim.equinox = obs->equinox;

		sprintf(lb, "%20.10f / ", fr->fim.xref);
		fits_add_keyword(fr, P_STR(FN_CRVAL1), lb);

		sprintf(lb, "%20.10f / ", fr->fim.yref);
		fits_add_keyword(fr, P_STR(FN_CRVAL2), lb);

		sprintf(lb, "%20.10f / ", fr->fim.rot);
		fits_add_keyword(fr, P_STR(FN_CROTA1), lb);

		if (obs->filter != NULL) {
			snprintf(lb, 127, "'%s'", obs->filter);
			fits_add_keyword(fr, P_STR(FN_FILTER), lb);
		}
		if (obs->objname != NULL) {
			snprintf(lb, 127, "'%s'", obs->objname);
			fits_add_keyword(fr, P_STR(FN_OBJECT), lb);
		}
		degrees_to_dms_pr(deg, obs->ra / 15.0, 2);
		sprintf(lb, "'%s'", deg);
		fits_add_keyword(fr, P_STR(FN_OBJCTRA), lb);

		degrees_to_dms_pr(deg, obs->dec, 1);
		sprintf(lb, "'%s'", deg);
		fits_add_keyword(fr, P_STR(FN_OBJCTDEC), lb);

		sprintf(lb, "%20.1f / EQUINOX OF COORDINATES", obs->equinox);
		fits_add_keyword(fr, P_STR(FN_EQUINOX), lb);
	}

	snprintf(lb, 127, "'%s'", P_STR(OBS_TELESCOP));
	fits_add_keyword(fr, P_STR(FN_TELESCOP), lb);

	snprintf(lb, 127, "'%s'", P_STR(OBS_FOCUS));
	fits_add_keyword(fr, P_STR(FN_FOCUS), lb);

	sprintf(lb, "%20.1f / TELESCOPE APERTURE IN CM", P_DBL(OBS_APERTURE));
	fits_add_keyword(fr, P_STR(FN_APERTURE), lb);

	sprintf(lb, "%20.1f / TELESCOPE FOCAL LENGTH IN CM", P_DBL(OBS_FLEN));
	fits_add_keyword(fr, P_STR(FN_FLEN), lb);

	snprintf(lb, 127, "'%s'", P_STR(OBS_OBSERVER));
	fits_add_keyword(fr, P_STR(FN_OBSERVER), lb);

	degrees_to_dms_pr(deg, P_DBL(OBS_LATITUDE), 1);
	sprintf(lb, "'%s'", deg);
	fits_add_keyword(fr, P_STR(FN_LATITUDE), lb);

	degrees_to_dms_pr(deg, P_DBL(OBS_LONGITUDE), 1);
	sprintf(lb, "'%s'", deg);
	fits_add_keyword(fr, P_STR(FN_LONGITUDE), lb);

	sprintf(lb, "%20.1f / OBSERVING SITE ALTITUDE (M)", P_DBL(OBS_ALTITUDE));
	fits_add_keyword(fr, P_STR(FN_ALTITUDE), lb);

}

/* read the wcs fields from the fits header lines
 * using parametrised field names */

void rescan_fits_wcs(struct ccd_frame *fr, struct wcs *fim)
{

	fim->wcsset = WCS_INITIAL;

	if (fits_get_double(fr, P_STR(FN_CRPIX1), &fim->xrefpix) <= 0)
		fim->wcsset = WCS_INVALID;
	if (fits_get_double(fr, P_STR(FN_CRPIX2), &fim->yrefpix) <= 0)
		fim->wcsset = WCS_INVALID;
	if (fits_get_double(fr, P_STR(FN_CRVAL1), &fim->xref) <= 0)
		fim->wcsset = WCS_INVALID;
	if (fits_get_double(fr, P_STR(FN_CRVAL2), &fim->yref) <= 0)
		fim->wcsset = WCS_INVALID;
	if (fits_get_double(fr, P_STR(FN_CDELT1), &fim->xinc) <= 0)
		fim->wcsset = WCS_INVALID;
	if (fits_get_double(fr, P_STR(FN_CDELT2), &fim->yinc) <= 0)
		fim->wcsset = WCS_INVALID;
	if (fits_get_double(fr, P_STR(FN_CROTA1), &fim->rot) <= 0)
		fim->rot = 0;
	if (fits_get_int(fr, P_STR(FN_EQUINOX), &fim->equinox) <= 0) {
		fim->equinox = 2000;
	}
	d3_printf("wcs status is: %d\n", fim->wcsset);
}

/* read the exp fields from the fits header lines
 * using parametrised field names */
void rescan_fits_exp(struct ccd_frame *fr, struct exp_data *exp)
{

	if (fits_get_int(fr, P_STR(FN_BINX), &exp->bin_x) > 0)
		exp->datavalid = 1;
	if (fits_get_int(fr, P_STR(FN_BINY), &exp->bin_y) > 0)
		exp->datavalid = 1;
	fits_get_double(fr, P_STR(FN_ELADU), &exp->scale);
	fits_get_double(fr, P_STR(FN_RDNOISE), &exp->rdnoise);
	fits_get_double(fr, P_STR(FN_FLNOISE), &exp->flat_noise);	
	if (fits_get_double(fr, P_STR(FN_DCBIAS), &exp->bias))
		exp->bias = 0;
	fits_get_double(fr, P_STR(FN_AIRMASS), &exp->airmass);
	if (fits_get_int(fr, P_STR(FN_SKIPX), &fr->x_skip) <= 0)
		fr->x_skip = 0;
	if (fits_get_int(fr, P_STR(FN_SKIPY), &fr->y_skip) <= 0)
		fr->y_skip = 0;
}

/* push the noise data from exp into the header fields of fr */
void noise_to_fits_header(struct ccd_frame *fr, struct exp_data *exp)
{
	char lb[128];
	sprintf(lb, "%20.3f / ELECTRONS / ADU", exp->scale);
	fits_add_keyword(fr, P_STR(FN_ELADU), lb);
	sprintf(lb, "%20.3f / READ NOISE IN ADUs", exp->rdnoise);
	fits_add_keyword(fr, P_STR(FN_RDNOISE), lb);
	sprintf(lb, "%20.3f / MULTIPLICATIVE NOISE COEFF", exp->flat_noise);
	fits_add_keyword(fr, P_STR(FN_FLNOISE), lb);
	sprintf(lb, "%20.3f / BIAS IN ADUs", exp->bias);
	fits_add_keyword(fr, P_STR(FN_DCBIAS), lb);
}


/* try to compute airmass from the frame header information 
 * return 0.0 if airmass couldn't be calculated */
double frame_airmass(struct ccd_frame *fr, double ra, double dec) 
{
	double v;
	double lat, lng, jd;
	char dms[64];
	double alt, az, airm;

	if (fits_get_double(fr, P_STR(FN_AIRMASS), &v) > 0) {
		return v;
	}

	if (fits_get_string(fr, P_STR(FN_LATITUDE), dms, 63) <= 0) 
		return 0.0;
	if (dms_to_degrees(dms, &lat))
		return 0.0;

	if (fits_get_string(fr, P_STR(FN_LONGITUDE), dms, 63) <= 0) 
		return 0.0;
	if (dms_to_degrees(dms, &lng))
		return 0.0;

	if (fits_get_double(fr, P_STR(FN_MJD), &v) > 0) {
		jd = v + 2500000.5;
	} else if (fits_get_double(fr, P_STR(FN_JDATE), &v) > 0) {
		jd = v;
	} else {
		return 0.0;
	}

	get_hrz_from_equ_sidereal_time (ra, dec, lng, lat,
					get_apparent_sidereal_time(jd),
					&alt, &az);

	airm = airmass(alt);

//	d3_printf("alt=%.4f az=%.4f sidereal=%.4f airmass=%.4f\n", alt, az, 
//		  get_apparent_sidereal_time(jd) - lng/15, airm);
	return airm;
}

double timeval_to_jdate(struct timeval *tv)
{
	int i, j, k;
	struct tm *t;
	int jdi;
	double jd;


	t = gmtime(&(tv->tv_sec));
// do the julian date (per hsaa p107)
	i = 1900 + t->tm_year;
	j = t->tm_mon + 1;
	k = t->tm_mday;
	jdi = k - 32075 + 1461 * (i + 4800 + (j - 14) / 12) / 4
		+ 367 * (j - 2 - (j - 14) / 12 * 12) / 12
		- 3 * (( i + 4900 + (j - 14) / 12) / 100) / 4;

	jd = jdi - 0.5 + t->tm_hour / 24.0 + t->tm_min / (24.0 * 60.0) 
		+ t->tm_sec / (24.0 * 3600.0) + tv->tv_usec / 
		(1000000.0 * 24.0 * 3600.0);
	return jd;
}



/* return the hour angle for the given obs coordinates, 
 * uses the parameter geo location values and current time */
double obs_current_hour_angle(struct obs_data *obs)
{
	double jd, ha;
	struct timeval tv;

	gettimeofday(&tv, NULL);
	jd = timeval_to_jdate(&tv);

	ha = get_apparent_sidereal_time(jd) - (P_DBL(OBS_LONGITUDE) + obs->ra) / 15.0;
	d3_printf("current ha = %f\n", ha);
	return ha;
}

/* return the airmass for the given obs coordinates, 
 * uses the parameter geo location values and current time */
double obs_current_airmass(struct obs_data *obs)
{
	double jd, airm;
	struct timeval tv;
	double alt, az;

	gettimeofday(&tv, NULL);
	jd = timeval_to_jdate(&tv);

	get_hrz_from_equ_sidereal_time (obs->ra, obs->dec, P_DBL(OBS_LONGITUDE), 
					P_DBL(OBS_LATITUDE),
					get_apparent_sidereal_time(jd),
					&alt, &az);
	airm = airmass(alt);
	d3_printf("current airmass = %f\n", airm);
	return airm;
}
