/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

/* create/read recipy files; read other star list files */

#define _GNU_SOURCE
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <libgen.h>
#include <glob.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "gcx.h"
#include "catalogs.h"
#include "gui.h"
#include "obsdata.h"
#include "sourcesdraw.h"
#include "params.h"
#include "interface.h"
#include "wcs.h"
#include "camera.h"
#include "cameragui.h"
#include "filegui.h"

#include "symbols.h"
#include "recipy.h"

#define INDENT 4 

/* names of symbols */
/* TODO: make sure everybody (reporting code in special) uses names from this table */
char *symname[SYM_LAST] = SYM_NAMES_INIT;

static inline void put_indent(int indent, int indstart, FILE *fp)
{
	int i;
	for (i = 0; i < indent * INDENT - indstart; i++)
		fputc(' ', fp);
}


static void recipy_report_head(FILE *fp, struct wcs *wcs, char *comment)
{
	char decs[64];
	char ras[64];

	fprintf(fp, "(");
	put_indent(1, 1, fp);
	if (wcs->wcsset < WCS_INITIAL) {
		err_printf("Bad wcs: skipping field center reporting in rcp\n");
		if (comment != NULL)
			fprintf(fp, "recipy (comments \"%s\")\n", comment);
		else
			fprintf(fp, "recipy ( )\n");
	} else {
		degrees_to_dms_pr(ras, wcs->xref / 15.0, 1);
		degrees_to_dms_pr(decs, wcs->yref, 0);
		fprintf(fp, "recipy (ra \"%s\" dec \"%s\" equinox %d",
			ras, decs, wcs->equinox);
		if ((comment != NULL) && (comment[0] != 0)) {
			fprintf(fp, "\n");
			put_indent(3, 0, fp);
			fprintf(fp, "comments \"%s\")\n", comment);
		} else {
			fprintf(fp, ")\n");
		}
	}
	put_indent(1, 0, fp);
	fprintf(fp, "stars (\n");
}

/* report a single cat star. return 1 if the star was reported successfully, 
 * 0 if it was skipped for some reason */
static int recipy_report_cat_star(FILE *fp, struct cat_star *cats, int indent, int type, 
				  double rcp_equinox)
{
	char ras[64], decs[64];
	char * typestr;

	if (cats == NULL)
		return 0;

	switch(type) {
	case MKRCP_STD:
		typestr = symname[SYM_STD];
		break;
	case MKRCP_TGT:
		typestr = symname[SYM_TARGET];
		break;
	case MKRCP_FIELD:
		typestr = symname[SYM_FIELD];
		break;
	case MKRCP_CAT:
		typestr = symname[SYM_CATALOG];
		break;
	default:
		d3_printf("ignoring star with unknown type %d in recipy_report_cat_star\n", 
			  type);
		return 0;
	}

	put_indent(indent, 2, fp);
	degrees_to_dms_pr(ras, cats->ra / 15.0, 2);
	degrees_to_dms_pr(decs, cats->dec, 1);
	if (cats->equinox != rcp_equinox) 
		fprintf(fp, "( name \"%s\" type %s ra \"%s\" dec \"%s\" equinox %.0f",
			cats->name, typestr, ras, decs, cats->equinox);
	else
		fprintf(fp, "( name \"%s\" type %s ra \"%s\" dec \"%s\"",
			cats->name, typestr, ras, decs);
	if (cats->comments != NULL && cats->comments[0] != 0) {
		fprintf(fp, "\n");
		put_indent(indent, 0, fp);
		fprintf(fp, "comments \"%s\"", cats->comments);
	}
	if (type != MKRCP_TGT) {
		fprintf(fp, "\n");
		put_indent(indent, 0, fp);
		if (cats->mag != 0.0)
			fprintf(fp, "mag %.1f smags \"%s\" ",
				cats->mag, cats->smags);
		else
			fprintf(fp, "smags \"%s\" ",
				cats->smags);
	} else {
		if (cats->mag != 0.0) {
			fprintf(fp, "\n");
			put_indent(indent, 0, fp);
			fprintf(fp, "mag %.1f ",
				cats->mag);
		} else if (cats->flags & (CAT_ASTROMET | CAT_VARIABLE)) {
			fprintf(fp, "\n");
			put_indent(indent, 0, fp);
		}
	}
	if (cats->flags & (CAT_ASTROMET | CAT_VARIABLE)) {
		fprintf(fp, "flags (");
		if (cats->flags & CAT_ASTROMET)
			fprintf(fp, " astrom");
		if (cats->flags & CAT_VARIABLE)
			fprintf(fp, " var");
		fprintf(fp, " ) )\n");
	} else {
		fprintf(fp, ")\n");
	}
	return 1;	
}


/* do the actual work of reporting the recipy. sl is the list of gui_stars 
 * return the number of stsr reported, or -1 for an error */
int report_recipy(FILE *rfp, GSList *rsl, struct wcs *wcs, 
			 int flags, char *comment)
{
	GSList *sl;
	struct gui_star *gs;
	struct cat_star cats;
	int ret = 0;
//	int indent = 1;

	recipy_report_head(rfp, wcs, comment);

	if (flags & MKRCP_STD) {
		sl = rsl;
		while (sl != NULL) {
			gs = GUI_STAR(sl->data);
			sl = g_slist_next(sl);
//			d3_printf("flags %x tmg %x tm %x\n", gs->flags,
//				  TYPE_MASK_GSTAR(gs), TYPE_MASK(STAR_TYPE_APSTD));
			if (((TYPE_MASK_GSTAR(gs) & TYPE_MASK(STAR_TYPE_APSTD)) == 0) 
			    || (gs->s == NULL))
				continue;
			ret += recipy_report_cat_star(rfp, CAT_STAR(gs->s), 2, MKRCP_STD, 
						      1.0 * wcs->equinox);
		}
	}
	if (flags & MKRCP_TGT) {
		sl = rsl;
		while (sl != NULL) {
			gs = GUI_STAR(sl->data);
			sl = g_slist_next(sl);
			if (((TYPE_MASK_GSTAR(gs) & TYPE_MASK(STAR_TYPE_APSTAR)) == 0)
			    || (gs->s == NULL))
				continue;
			ret += recipy_report_cat_star(rfp, CAT_STAR(gs->s), 2, MKRCP_TGT,
						      1.0 * wcs->equinox);
		}
	}
 	if (flags & MKRCP_CAT) {
		sl = rsl;
		while (sl != NULL) {
			gs = GUI_STAR(sl->data);
			sl = g_slist_next(sl);
			if (((TYPE_MASK_GSTAR(gs) & TYPE_MASK(STAR_TYPE_CAT)) == 0)
			     || (gs->s == NULL))
				continue;
			if (flags & MKRCP_CAT_TO_STD)
				ret += recipy_report_cat_star(rfp, CAT_STAR(gs->s), 2, MKRCP_STD,
							      1.0 * wcs->equinox);
			else
				ret += recipy_report_cat_star(rfp, CAT_STAR(gs->s), 2, MKRCP_CAT,
							      1.0 * wcs->equinox);
		}
	}
 	if (flags & MKRCP_FIELD) {
		sl = rsl;
		while (sl != NULL) {
			gs = GUI_STAR(sl->data);
			sl = g_slist_next(sl);
			if (((TYPE_MASK_GSTAR(gs) & (TYPE_MASK(STAR_TYPE_SREF) )) == 0)
			     || (gs->s == NULL))
				continue;
			if (flags & MKRCP_FIELD_TO_TGT)
				ret += recipy_report_cat_star(rfp, CAT_STAR(gs->s), 2, MKRCP_TGT,
							      1.0 * wcs->equinox);
			else
				ret += recipy_report_cat_star(rfp, CAT_STAR(gs->s), 2, MKRCP_FIELD,
							      1.0 * wcs->equinox);
		}
	}
 	if (flags & MKRCP_USER) {
		sl = rsl;
		while (sl != NULL) {
			double ra, dec;
			gs = GUI_STAR(sl->data);
			sl = g_slist_next(sl);
			if ((TYPE_MASK_GSTAR(gs) & TYPE_MASK(STAR_TYPE_USEL)) == 0)
				continue;
			w_worldpos(wcs, gs->x, gs->y, &ra, &dec);
			snprintf(cats.name, CAT_STAR_NAME_SZ, "x%.0fy%.0f", gs->x,  gs->y);
			cats.ra = ra;
			cats.dec = dec;
			cats.equinox = 1.0 * wcs->equinox;
			cats.mag = 0.0;
			cats.flags = 0;
			cats.comments = NULL;
			if (flags & MKRCP_USER_TO_TGT)
				ret += recipy_report_cat_star(rfp, &cats, 2, MKRCP_TGT,
							      1.0 * wcs->equinox);
			else
				ret += recipy_report_cat_star(rfp, &cats, 2, MKRCP_FIELD,
							      1.0 * wcs->equinox);
		}
	}
 	if (flags & MKRCP_DET) {
		sl = rsl;
		while (sl != NULL) {
			double ra, dec;
			gs = GUI_STAR(sl->data);
			sl = g_slist_next(sl);
			if ((TYPE_MASK_GSTAR(gs) & TYPE_MASK(STAR_TYPE_SIMPLE)) == 0)
				continue;
			w_worldpos(wcs, gs->x, gs->y, &ra, &dec);
			snprintf(cats.name, CAT_STAR_NAME_SZ, "x%.0fy%.0f", gs->x,  gs->y);
			cats.ra = ra;
			cats.dec = dec;
			cats.equinox = 1.0 * wcs->equinox;
			cats.mag = 0.0;
			cats.flags = 0;
			cats.comments = NULL;
			if (flags & MKRCP_DET_TO_TGT)
				ret += recipy_report_cat_star(rfp, &cats, 2, MKRCP_TGT,
							      1.0 * wcs->equinox);
			else
				ret += recipy_report_cat_star(rfp, &cats, 2, MKRCP_FIELD,
							      1.0 * wcs->equinox);
		}
	}

	put_indent(1, 0, rfp);
	fprintf(rfp, ")\n)\n");
	fflush(rfp);

	return ret;
}


/* recipy reading */

/* create a scanner and tell it about the symbols we look for */
GScanner* init_scanner()
{
	int i;
	GScanner *scan;

	scan = g_scanner_new(NULL);
	if (scan == NULL)
		return scan;
	scan->config->cpair_comment_single = ";\n";
	scan->config->int_2_float = 1;
	scan->config->case_sensitive = 0;
//	scan->config->scan_symbols = 0;

	for (i=1; i<SYM_LAST; i++) {
		g_scanner_add_symbol(scan, symname[i], (gpointer)i);
	}
	return scan;
}

/* print the token in a friendly form */
static void print_token(GScanner * scan, GTokenType tok)
{
	if (tok < G_TOKEN_NONE)
		d3_printf("tok: %c\n", tok);
	else {
		if (tok == G_TOKEN_STRING)
			d3_printf("tok: string val: %s\n",  
				  g_scanner_cur_value(scan).v_string);
		else if (tok == G_TOKEN_IDENTIFIER)
			d3_printf("tok: ident val: %s\n",  
				  g_scanner_cur_value(scan).v_string);
		else if (tok == G_TOKEN_SYMBOL)
			d3_printf("tok: symbol [%s] val: %d\n", 
				  symname[g_scanner_cur_value(scan).v_int],
				  (int)g_scanner_cur_value(scan).v_int);
		else if (tok == G_TOKEN_INT)
			d3_printf("tok: int val: %d\n", 
				  (int)g_scanner_cur_value(scan).v_int);
		else if (tok == G_TOKEN_FLOAT)
			d3_printf("tok: float val: %f\n", 
				  g_scanner_cur_value(scan).v_float);
		else
			d3_printf("tok: %d\n", tok);
	}
}


void test_parse_rcp(void)
{
	GScanner *scan;
	FILE *fp;
	GTokenType tok;

	fp = fopen("test.rcp", "r");
	if (fp == NULL) {
		err_printf("cannot open test.rcp\n");
		return;
	}

	scan = init_scanner();
	g_scanner_input_file(scan, fileno(fp));
	do {
		tok = g_scanner_get_next_token(scan);
		print_token(scan, tok);
	} while (tok != G_TOKEN_EOF);
	g_scanner_destroy(scan);

}



/* skip until we get a closing brace */

static void skip_list(GScanner *scan) 
{
	int nbr = 1;
	int tok;
	do {
		tok = g_scanner_get_next_token(scan);
//		d3_printf("skipping ");
//		print_token(scan, tok);
		if (tok == '(')
			nbr ++;
		else if (tok == ')')
			nbr --;
	} while ((tok != G_TOKEN_EOF) && (nbr > 0));
}

/* skip one item (or a whole list if that is the case */
static void skip_one(GScanner *scan) 
{
	int tok;
	tok = g_scanner_get_next_token(scan);
//	d3_printf("skipping ");
//	print_token(scan, tok);
	if (tok == '(') 
		skip_list(scan);
}


/* read the flags from the list (open brace has already been read) 
 * stop when the closing brace is found, or on eof */
static int get_flags_from_list(GScanner *scan)
{
	GTokenType tok;
	int flags = 0;

	do {
		tok = g_scanner_get_next_token(scan);
//		d3_printf("flags ");
//		print_token(scan, tok);
		if (tok == '(') {
			skip_list(scan);
			continue;
		}
		if (tok != G_TOKEN_SYMBOL)
			continue;
		switch(intval(scan)) {
		case SYM_ASTROM:
			flags |= CAT_ASTROMET;
			break;
		case SYM_VAR:
			flags |= CAT_VARIABLE;
			break;
		case SYM_CENTERED:
			flags |= CPHOT_CENTERED;
			break;
		case SYM_NOT_FOUND:
			flags |= CPHOT_NOT_FOUND;
			break;
		case SYM_UNDEF_ERR:
			flags |= CPHOT_UNDEF_ERR;
			break;
		case SYM_HAS_BADPIX:
			flags |= CPHOT_BADPIX;
			break;
		case SYM_BRIGHT:
			flags |= CPHOT_BURNED;
			break;
		case SYM_NO_COLOR:
			flags |= CPHOT_NO_COLOR;
			break;
		default:
			break;
		}
	} while (tok != ')' && tok != G_TOKEN_EOF);
//	d3_printf("flags is %x\n", flags / 16);
	return flags;
}

/* read the star, up to the final closing bracket. return 0 if we could
 * find at least a ra/dec and a name */

int parse_star(GScanner *scan, struct cat_star *cats)
{
	int havera = 0, havedec = 0, havename = 0, havemag = 0;
	GTokenType tok;

	do {
		tok = g_scanner_get_next_token(scan);
//		d3_printf("star ");
//		print_token(scan, tok);
		if (tok == ')')
			continue;
		if (tok != G_TOKEN_SYMBOL) {
//			d3_printf("skip non-symbol\n");
			if (tok == '(')
				skip_list(scan);
			else 
				skip_one(scan);
			continue;
		}
		switch(intval(scan)) {
		case SYM_RA:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case G_TOKEN_STRING:
				havera = !dms_to_degrees(stringval(scan),
							  &cats->ra);
				cats->ra *= 15.0;
				break;
			case '-':
				tok = g_scanner_get_next_token(scan);
				if (tok != G_TOKEN_FLOAT)
					break;
				cats->ra = -floatval(scan);
				havera = 1;
				break;
			case G_TOKEN_FLOAT:
				cats->ra = floatval(scan);
				havera = 1;
				break;
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		case SYM_DEC:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case G_TOKEN_STRING:
				havedec = !dms_to_degrees(stringval(scan),
							  &cats->dec);
				break;
			case '-':
				tok = g_scanner_get_next_token(scan);
				if (tok != G_TOKEN_FLOAT)
					break;
				cats->dec = -floatval(scan);
				havedec = 1;
				break;
			case G_TOKEN_FLOAT:
				cats->dec = floatval(scan);
				havedec = 1;
				break;
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		case SYM_EQUINOX:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case '-':
				tok = g_scanner_get_next_token(scan);
				if (tok != G_TOKEN_FLOAT)
					break;
				cats->equinox = -floatval(scan);
				break;
			case G_TOKEN_FLOAT:
				cats->equinox = floatval(scan);
				break;
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		case SYM_MAG:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case '-':
				tok = g_scanner_get_next_token(scan);
				if (tok != G_TOKEN_FLOAT)
					break;
				cats->mag = -floatval(scan);
				havemag = 1;
				break;
			case G_TOKEN_FLOAT:
				cats->mag = floatval(scan);
				havemag = 1;
				break;
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		case SYM_RESIDUAL:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case '-':
				tok = g_scanner_get_next_token(scan);
				if (tok != G_TOKEN_FLOAT)
					break;
				cats->residual = -floatval(scan);
				break;
			case G_TOKEN_FLOAT:
				cats->residual = floatval(scan);
				break;
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		case SYM_WEIGHT:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case '-':
				tok = g_scanner_get_next_token(scan);
				if (tok != G_TOKEN_FLOAT)
					break;
				cats->weight = -floatval(scan);
				break;
			case G_TOKEN_FLOAT:
				cats->weight = floatval(scan);
				break;
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		case SYM_COMMENTS:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case G_TOKEN_STRING:
				cats->comments = strdup(stringval(scan));
				break;
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		case SYM_SMAGS:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case G_TOKEN_STRING:
				cats->smags = strdup(stringval(scan));
				break;
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		case SYM_IMAGS:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case G_TOKEN_STRING:
				cats->imags = strdup(stringval(scan));
				break;
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		case SYM_NAME:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case G_TOKEN_STRING:
				strncpy(cats->name, stringval(scan), CAT_STAR_NAME_SZ);
				havename = 1;
				break;
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		case SYM_TYPE:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case G_TOKEN_SYMBOL:
				switch(intval(scan)) {
				case SYM_STD:
					cats->flags = (cats->flags & ~CAT_STAR_TYPE_MASK)
						| CAT_STAR_TYPE_APSTD;
					cats->flags &= ~CAT_VARIABLE;
					break;
				case SYM_TGT:
				case SYM_TARGET:
					cats->flags = (cats->flags & ~CAT_STAR_TYPE_MASK)
						| CAT_STAR_TYPE_APSTAR;
					break;
				case SYM_FIELD:
					cats->flags = (cats->flags & ~CAT_STAR_TYPE_MASK)
						| CAT_STAR_TYPE_SREF;
					break;
				case SYM_CATALOG:
					cats->flags = (cats->flags & ~CAT_STAR_TYPE_MASK)
						| CAT_STAR_TYPE_CAT;
					break;
				default:
					break;
				}
				break;
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		case SYM_FLAGS:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case '(':
				cats->flags |= get_flags_from_list(scan);
				break;
			default:
				break;
			}
			break;
		default:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		}
	} while (tok != ')' && tok != G_TOKEN_EOF);
	if (!havemag) {
//		mag_from_smags(cats);
	}
	if (havedec && havera && havename)
		return 0;
	else
		return -1;
}



/* read a recipy an return the number of stars read. Place pointers to newly-allocated
 * cat_stars in cst. Fill in the ra and dec of the field (if possible) 
 * return -1 for fatal errors, otherwise try to scan as much as possible */

static int read_new_rcp(struct cat_star *cst[], FILE *fp, int n, double *ra, double *dec)
{
	GScanner *scan;
	GTokenType tok;
	RcpState state = RCP_STATE_START;

	double ira, idec, iequ = 2000.0;
	int haveira=0, haveidec=0;

	int i = 0;
	struct cat_star *cats;

	scan = init_scanner();
	g_scanner_input_file(scan, fileno(fp));
	if (g_scanner_peek_next_token(scan) != '(')
		return -1; /* must be old-style recipy */
	do {
		tok = g_scanner_get_next_token(scan);
//		d3_printf("state %d ", state);
//		print_token(scan, tok);
		switch(state) {
		case RCP_STATE_START:
			if (tok == '(')
				state = RCP_STATE_TOPALIST;
			break; /* we just skip stuff up to the first '(' */
		case RCP_STATE_TOPALIST:
			switch(tok) {
			case G_TOKEN_SYMBOL:
				if ((intval(scan) == SYM_RECIPY) 
				    || (intval(scan) == SYM_OBSERVATION)) {
					tok = g_scanner_get_next_token(scan);
					if (tok != '(') 
						break;
					state = RCP_STATE_RCPALIST;
				} else if (intval(scan) == SYM_STARS){
					tok = g_scanner_get_next_token(scan);
					if (tok != '(') 
						break;
					state = RCP_STATE_STARLIST;
				} else {
					skip_one(scan);
				}
				break;
			case ')':
				state = RCP_STATE_END;
				break;
			case '(':
				skip_list(scan);
				break;
			default: /* print a warning here */
				state = RCP_STATE_SKIP;
					break;
			}
			break;
		case RCP_STATE_RCPALIST: 
			switch(tok) {
			case ')':
				state = RCP_STATE_TOPALIST;
				break;
			case '(':
				skip_list(scan);
				break;
			case G_TOKEN_SYMBOL:
				switch(intval(scan)) {
				case SYM_RA:
					tok = g_scanner_get_next_token(scan);
					switch(tok) {
					case G_TOKEN_STRING:
						haveira = !dms_to_degrees(stringval(scan),
									  &ira);
						ira *= 15.0;
						break;
					case G_TOKEN_FLOAT:
						ira = floatval(scan);
						haveira = 1;
						break;
					case '(':
						skip_list(scan);
						break;
					case ')':
						state = RCP_STATE_TOPALIST;
						break;
					default:
						break;
					}
					break;
				case SYM_DEC:
					tok = g_scanner_get_next_token(scan);
					switch(tok) {
					case G_TOKEN_STRING:
						haveidec = !dms_to_degrees(stringval(scan),
									  &idec);
						break;
					case '-':
						tok = g_scanner_get_next_token(scan);
						if (tok != G_TOKEN_FLOAT)
							break;
						idec = -floatval(scan);
						haveidec = 1;
						break;
					case G_TOKEN_FLOAT:
						idec = floatval(scan);
						haveidec = 1;
						break;
					case '(':
						skip_list(scan);
						break;
					case ')':
						state = RCP_STATE_TOPALIST;
						break;
					default:
						break;
					}
					break;
				case SYM_EQUINOX:
					tok = g_scanner_get_next_token(scan);
					switch(tok) {
					case G_TOKEN_FLOAT:
						iequ = floatval(scan);
						break;
					case '(':
						skip_list(scan);
						break;
					case ')':
						state = RCP_STATE_TOPALIST;
						break;
					default:
						break;
					}
					break;
				default:
					tok = g_scanner_get_next_token(scan);
					switch(tok) {
					case '(':
						skip_list(scan);
						break;
					case ')':
						state = RCP_STATE_TOPALIST;
						break;
					default:
						break;
					}
					break;
				}
			default:
				break;
			}
			break;
		case RCP_STATE_STARLIST:
			if (tok == '(') {
				if (i >= n) {
					skip_list(scan);
					break;
				}
				cats = cat_star_new();
				if (cats == NULL) {
					skip_list(scan);
					break;
				}
				cats->equinox = iequ;
				if (!parse_star(scan, cats)) {
					cst[i] = cats;
					i++;
				} else {
					cat_star_release(cats);
				}
//				d3_printf("end of star\n");
				break;
			} else if (tok == ')') {
				state = RCP_STATE_TOPALIST;
				break;
			} else {
				break;
			}
			break;
		default:
			break;
		}
	} while (tok != G_TOKEN_EOF);
	g_scanner_destroy(scan);
	return i;

}

/* old style recypy reading */

/* parse a 'constar' line into the cats. Return 0 if successfull */
static int crack_constar(char *line, struct cat_star *cats)
{
	char comment[256];
	char name[256], *smags = NULL;
	float ra, dec, mag = 0.0, vmag=0.0;
	int ret, i;

	ret = sscanf(line+7, " %255s %f %f PMAG=%f", 
		     name, &ra, &dec, &mag);
	if (ret < 4)
		return -1;

	comment[0] = 0;
	i=0;
	while (line[i] != 0) {
		if (!strncasecmp(line + i, "VMAG", 4))
			break;
		i++;
	}
	if (line[i] == 0) {
/* field star */
		ret = asprintf(&smags, "p(gsc)=%.3f", mag);
		if (ret < 0)
			smags = NULL;
		cats->flags = CAT_ASTROMET | CAT_STAR_TYPE_SREF;
	} else {
/* ref star (has vmag) */
		ret = sscanf(line+i, "VMAG=%f", &vmag);
		if (ret == 1) {
			ret = asprintf(&smags, "v(aavso)=%.3f p(gsc)=%.3f", vmag, mag);
			if (ret < 0)
				smags = NULL;
		}
		cats->flags = CAT_STAR_TYPE_APSTD | CAT_ASTROMET;
	}
	i=0;
	while (line[i] != 0) {
		if (!strncasecmp(line + i, "GSCL=", 5))
			break;
		i++;
	}
	if (line[i] != 0) {
		ret = sscanf(line+i, "GSCL=%255s", name);
	}

	cats->ra = ra;
	cats->dec = dec;
	cats->equinox = 2000.0;
	cats->mag = mag;
	cats->smags = smags;
	strncpy(cats->name, name, CAT_STAR_NAME_SZ);
//	cats->flags = CAT_PHOTOMET;

//	d3_printf("constar name:%s ra:%.4f dec:%.4f mag:%.4f smags:%s\n", 
//		  cats->name, cats->ra, cats->dec, cats->mag, cats->smags);
	return 0;
}

/* parse a 'varstar' line into the cats. Return 0 if successfull */
static int crack_varstar(char *line, char *line2, struct cat_star *cats)
{
	char comment[256];
	char name[256];
	float ra, dec, mag = 0.0;
	int ret, i, k = 0;

	ret = sscanf(line+7, " %255s %f %f %*cMAG=(%f", 
		     name, &ra, &dec, &mag);
	if (ret < 3)
		return -1;

	comment[0] = 0;
	i=0;
	while (line[i] != 0) {
		if (!strncasecmp(line + i, "MAG", 3))
			break;
		i++;
	}
	if (line[i] == 0) 
		goto end;
	if (i > 1)
		i--;

	while (isprint(line[i]) && k < 255) {
		comment[k++] = line[i++];
	}
	comment[k] = 0;
	if (line2 != NULL) {
		i=0;
		while (line2[i] != 0) {
			if (!strncasecmp(line2 + i, "VARID", 5))
				break;
			i++;
		}
		if (line2[i] == 0) 
			goto end;
		while (isprint(line2[i]) && line2[i] != ' ' && k < 255) {
			comment[k++] = line2[i++];
		}
		comment[k] = 0;
	}

end:
	cats->ra = ra;
	cats->dec = dec;
	cats->equinox = 2000.0;
	cats->mag = mag;
	strncpy(cats->name, name, CAT_STAR_NAME_SZ);
	cats->comments = strndup(comment, k);
	cats->flags = CAT_STAR_TYPE_APSTAR | CAT_VARIABLE;

//	d3_printf("varstar name:%s ra:%.4f dec:%.4f mag:%.4f k=%d\ncomment:%s\n", 
//		  cats->name, cats->ra, cats->dec, cats->mag, k, cats->comments);
	return 0;
}

/* read a avsomat-style rcp file and return it as 
 * at most n cat_stars 
 * in the supplied table. Return the number of stars found
 * or a negative error. name is the file name */
int read_avsomat_rcp(struct cat_star *cst[], FILE *fp, int n)
{
	struct cat_star *cats = NULL;
	FILE *rfn = fp;
	int sn = 0, linesz, line2sz;
	char *line, *line2;

	if (cst == NULL)
		return -1;

	while ( (sn < n)) {
		line = NULL;
		linesz = 0;
		if (getline(&line, &linesz, rfn) < 0)
			break;
//		d3_printf("line is %s\n", line);
		if (cats == NULL)
			cats = cat_star_new();
		if (cats == NULL) {
			free(line);
			return -1;
		}
		if (!strncasecmp(line, "VARSTAR", 7)) {
			if (cats == NULL) {
				free(line);
				return -1;
			}
			line2 = NULL;
			line2sz = 0;
			if (getline(&line2, &line2sz, rfn) < 0) {
				line2 = NULL;
			}
//			d3_printf("line2 is %s\n", line2);
			if (!crack_varstar(line, line2, cats)) {
				cst[sn] = cats;
				sn ++;
				cats = NULL;
			} 
			free(line);
			if (line2)
				free(line2);
			continue;
		}
		if (!strncasecmp(line, "CONSTAR", 7)) {
			if (!crack_constar(line, cats)) {
				cst[sn] = cats;
				sn ++;
				cats = NULL;
			} 
			continue;
		}
	}
	if (cats != NULL)
		cat_star_release(cats);
	return sn;
}

/* read a recipy (old or new), and return the number of stars read, or a negative 
 * error; update ra and dec to the field center coordinates */
int read_rcp(struct cat_star *cst[], FILE *fp, int n, double *ra, double *dec)
{
	int ret;
	double ira = 0.0, idec = 0.0;

	ret = read_new_rcp(cst, fp, n, &ira, &idec);
	if (ret < 0)
		ret = read_avsomat_rcp(cst, fp, n);
	if (ret < 0)
		return ret;
	if (ira == 0.0 && idec == 0.0) {
/* get the ra and dec of the first star as field center */
		ira = cst[0]->ra;
		idec = cst[0]->dec;
	}
	if (ra)
		*ra = ira;
	if (dec)
		*dec = idec;
	return ret;
}

/* convert an old style recipy to new. Return the number of stars written */
#define MAX_STARS_CONV 65536
int convert_recipy(FILE *inf, FILE *outf) 
{
	struct cat_star *csl[MAX_STARS_CONV];
	int ret = 0, n, i;
	double ra, dec;
	struct wcs wcs; /* dummy wcs for the header */

	n = read_rcp(csl, inf, MAX_STARS_CONV, &ra, &dec);
	if (n <= 0)
		return n;

	wcs.xref = ra;
	wcs.yref = dec;
	wcs.equinox = 2000.0;

	recipy_report_head(outf, &wcs, NULL);

	for (i=0; i<n; i++) {
		if (CATS_TYPE(csl[i]) == CAT_STAR_TYPE_APSTD)
			ret += recipy_report_cat_star(outf, csl[i], 2, MKRCP_STD, 
						      1.0 * wcs.equinox);
	}
	for (i=0; i<n; i++) {
		if (CATS_TYPE(csl[i]) == CAT_STAR_TYPE_APSTAR)
			ret += recipy_report_cat_star(outf, csl[i], 2, MKRCP_TGT, 
						      1.0 * wcs.equinox);
	}
	for (i=0; i<n; i++) {
		if (CATS_TYPE(csl[i]) == CAT_STAR_TYPE_SREF)
			ret += recipy_report_cat_star(outf, csl[i], 2, MKRCP_FIELD, 
						      1.0 * wcs.equinox);
	}
	for (i=0; i<n; i++) {
		cat_star_release(csl[i]);
	}
	put_indent(1, 0, outf);
	fprintf(outf, ")\n)\n");
	fflush(outf);
	return ret;
} 

static int mag_comp_fn (const void *a, const void *b)
{
	struct cat_star **ca = (struct cat_star **) a; 
	struct cat_star **cb = (struct cat_star **) b; 
	return ((*ca)->mag > (*cb)->mag) - ((*ca)->mag < (*cb)->mag);
}

#define MAX_STARS_READ 100000
/* load stars from a donwloaded gsc2 text file; return the number of stars read 
 * or -1 for an error; */
int read_gsc2(struct cat_star *csl[], FILE *fp, int n, double *cra, double *cdec)
{
	struct cat_star *cats = NULL;
	struct cat_star *cst[MAX_STARS_READ];
	FILE *rfn = fp;
	int sn = 0, linesz, i;
	char *line;
	float ra, dec, epoch, fmag, fmagerr, jmag, jmagerr, vmag, vmagerr, nmag, nmagerr;
	char id[64];
	int ret;

	if (csl == NULL)
		return -1;

	line = NULL;
	linesz = 0;

	while ( (sn < MAX_STARS_READ)) {
		if (getline(&line, &linesz, rfn) < 0)
			break;

		ret = sscanf(line, "%63s %f %f %*f %*f %f %*f %*f %*f %*f %f %f %f %f %f %f %f %f",
			     id, &ra, &dec, &epoch, &fmag, &fmagerr, &jmag, &jmagerr, 
			     &vmag, &vmagerr, &nmag, &nmagerr);
		if (ret != 12) 
			continue;

		cats = cat_star_new();
		if (cats == NULL) {
			free(line);
			return -1;
		}

		strncpy(cats->name, id, CAT_STAR_NAME_SZ);
		cats->ra = ra;
		cats->dec = dec;
		if (vmag < 30)
			cats->mag = vmag;
		else if (fmag < 30)
			cats->mag = fmag;
		else if (jmag < 30)
			cats->mag = jmag;
		cats->equinox = 2000.0;
		if (vmag < 30) {
			update_band_by_name(&cats->smags, "v(gsc2)", vmag, vmagerr);
		}
		if (fmag < 30) {
			update_band_by_name(&cats->smags, "f(gsc2)", fmag, fmagerr);
		}
		if (jmag < 30) {
			update_band_by_name(&cats->smags, "j(gsc2)", jmag, jmagerr);
		}
		if (nmag < 30) {
			update_band_by_name(&cats->smags, "n(gsc2)", nmag, nmagerr);
		}
		cats->flags = CAT_STAR_TYPE_SREF | CAT_ASTROMET;
		asprintf(&cats->comments, "p=g ");
		cst[sn]=cats;
		sn ++;
	}
	if (sn > 2) {
		qsort(cst, sn, sizeof(struct cat_star *), mag_comp_fn);
	}
	for (i=0; i<n && i<sn; i++)
		csl[i] = cst[i];
	for (; i<sn; i++) 
		cat_star_release(cst[i]);
	return n < sn ? n : sn;
}



#define MAX_STARS_READ 100000
/* read a landolt table (only used to convert) */
int read_landolt_table(struct cat_star *csl[], FILE *fp, int n, double *cra, double *cdec)
{
	struct cat_star *cats = NULL;
	struct cat_star *cst[MAX_STARS_READ];
	FILE *rfn = fp;
	int sn = 0, linesz, i;
	char *line;
	float vmag, bmag, umag, rmag, imag;
	double ra, dec;
	char id[64];
	char nam[64];
	char ras[64], decs[64];
	int ret, len;

	if (csl == NULL)
		return -1;

	line = NULL;
	linesz = 0;

	while ( (sn < MAX_STARS_READ)) {
		if (getline(&line, &linesz, rfn) < 0)
			break;

		if ((len = strlen(line)) < 48)
			continue;
		line[13] = ':';
		line[16] = ':';
		line[25] = ':';
		line[28] = ':';

		ret = sscanf(line, "%s %s %s", id, ras, decs);
		if (ret != 3)
			continue;
		ret = sscanf(line+46, "%f %*f %f %*f %f %*f %f %*f %f %*f %[^\n]",  
			     &vmag, &bmag, &umag, &rmag, &imag, nam);
		if (ret != 6) 
			continue;

		if (dms_to_degrees(ras, &ra))
			continue;
		if (dms_to_degrees(decs, &dec))
			continue;

		cats = cat_star_new();
		if (cats == NULL) {
			free(line);
			return -1;
		}

		strncpy(cats->name, id, CAT_STAR_NAME_SZ);
		cats->ra = ra * 15.0;
		cats->dec = dec;
		cats->mag = vmag;
		cats->equinox = 2000.0;
		asprintf(&cats->smags, "v(landolt)=%.3f/0.005 b(landolt)=%.3f/0.005 "
			 "u(landolt)=%.3f/0.005 r(landolt)=%.3f/0.005 i(landolt)=%.3f/0.005",
			 vmag, bmag, umag, rmag, imag);
		cats->flags = CAT_STAR_TYPE_APSTD | CAT_ASTROMET;
		asprintf(&cats->comments, "%s", nam);
		cst[sn]=cats;
		sn ++;
	}
	for (i=0; i<n && i<sn; i++)
		csl[i] = cst[i];
	for (; i<sn; i++) 
		cat_star_release(cst[i]);
	return n < sn ? n : sn;
}


#define MAX_STARS_READ 100000
/* read a landolt table (only used to convert) */
int read_henden_table(struct cat_star *csl[], FILE *fp, int n, double *cra, double *cdec)
{
	struct cat_star *cats = NULL;
	struct cat_star *cst[MAX_STARS_READ];
	FILE *rfn = fp;
	int sn = 0, linesz, i, nc;
	char *line;
	float vmag, bvmag, ubmag, vrmag, rimag;
	float verr, bverr, uberr, vrerr, rierr;
	float ra, dec, raerr, decerr;
	char id[64];
	int ret, len;

	if (csl == NULL)
		return -1;

	line = NULL;
	linesz = 0;

	while ( (sn < MAX_STARS_READ)) {
		if (getline(&line, &linesz, rfn) < 0)
			break;

		if ((len = strlen(line)) < 48)
			continue;

		d3_printf("reading: %s\n", line);

		ret = sscanf(line, "%s %f %f %f %f", id, &ra, &raerr, &dec, &decerr);
		if (ret != 5)
			continue;
		ret = sscanf(line+46, "%*f %f %f %f %f %f %f %f %f %f %f",  
			     &vmag, &bvmag, &ubmag, &vrmag, &rimag,
			     &verr, &bverr, &uberr, &vrerr, &rierr);
		if (ret != 10) 
			continue;

		d3_printf("vmag=%f\n", vmag);

		cats = cat_star_new();
		if (cats == NULL) {
			free(line);
			return -1;
		}

		snprintf(cats->name, CAT_STAR_NAME_SZ, "%s_%d", id, sn+1);
		cats->ra = ra;
		cats->ra_err = raerr;
		cats->dec = dec;
		cats->dec_err = decerr;
		cats->mag = vmag;
		cats->equinox = 2000.0;
		cats->smags = calloc(1, 256);
		if (!cats->smags)
			continue;
		i = 0;
		nc = 255;
		if (vmag < 90.0 && verr < 9) {
			i += snprintf(cats->smags + i, nc-i, 
				      "v(henden)=%.3f/%.3f",
				      vmag, verr);
		}
		if (bvmag < 90.0 && bverr < 9 && i < nc) {
			i += snprintf(cats->smags + i, nc-i, 
				      " b-v(henden)=%.3f/%.3f",
				      bvmag, bverr);
		}
		if (ubmag < 90.0 && uberr < 9 && i < nc) {
			i += snprintf(cats->smags + i, nc-i, 
				      " u-b(henden)=%.3f/%.3f",
				      ubmag, uberr);
		}
		if (vrmag < 90.0 && vrerr < 9 && i < nc) {
			i += snprintf(cats->smags + i, nc-i, 
				      " v-r(henden)=%.3f/%.3f",
				      vrmag, vrerr);
		}
		if (rimag < 90.0 && rierr < 9 && i < nc) {
			i += snprintf(cats->smags + i, nc-i, 
				      " r-i(henden)=%.3f/%.3f",
				      rimag, rierr);
		}
		cats->flags = CAT_STAR_TYPE_APSTD | CAT_ASTROMET;
		cst[sn]=cats;
		sn ++;
	}
	for (i=0; i<n && i<sn; i++)
		csl[i] = cst[i];
	for (; i<sn; i++) 
		cat_star_release(cst[i]);
	return n < sn ? n : sn;
}


/* convert a catalog table file to recipy. Return the number of stars written 
 * this function works with a particular file format only */
#define MAX_STARS_CONV 65536
int convert_catalog(FILE *inf, FILE *outf, char *catname) 
{
	struct cat_star *csl[MAX_STARS_CONV];
	int ret = 0, n = -1, i;
	double ra, dec;
	struct wcs wcs; /* dummy wcs for the header */

	if (!strcasecmp(catname, "landolt")) {
		n = read_landolt_table(csl, inf, MAX_STARS_CONV, &ra, &dec);
	} else if (!strcasecmp(catname, "henden")) {
		n = read_henden_table(csl, inf, MAX_STARS_CONV, &ra, &dec);
	} else if (!strcasecmp(catname, "gcvs")) {
		err_printf("no gcvs yet\n");
	} else {
		return -1;
	}
	if (n <= 0)
		return n;

	wcs.xref = 0;
	wcs.yref = 0;
	wcs.equinox = 2000.0;

	recipy_report_head(outf, &wcs, NULL);

	for (i=0; i<n; i++) {
		if (CATS_TYPE(csl[i]) == CAT_STAR_TYPE_APSTD)
			ret += recipy_report_cat_star(outf, csl[i], 2, MKRCP_STD, 
						      1.0 * wcs.equinox);
	}
	for (i=0; i<n; i++) {
		if (CATS_TYPE(csl[i]) == CAT_STAR_TYPE_APSTAR)
			ret += recipy_report_cat_star(outf, csl[i], 2, MKRCP_TGT, 
						      1.0 * wcs.equinox);
	}
	for (i=0; i<n; i++) {
		if (CATS_TYPE(csl[i]) == CAT_STAR_TYPE_SREF)
			ret += recipy_report_cat_star(outf, csl[i], 2, MKRCP_FIELD, 
						      1.0 * wcs.equinox);
	}
	for (i=0; i<n; i++) {
		cat_star_release(csl[i]);
	}
	put_indent(1, 0, outf);
	fprintf(outf, ")\n)\n");
	fflush(outf);
	return ret;
} 

static int parse_obs(GScanner *scan, struct obs_data *obs);

int read_obs_report(FILE *fp, struct cat_star *cst[], int n, 
			   struct obs_data *obst[], int no)
{
	GScanner *scan;
	GTokenType tok;
	RcpState state = RCP_STATE_START;

	struct obs_data *obs = NULL;

	int i = 0, io = 0;
	struct cat_star *cats;
	int ret;

	scan = init_scanner();
	g_scanner_input_file(scan, fileno(fp));
	if (g_scanner_peek_next_token(scan) != '(')
		return -1; /* must be old-style recipy */
	do {
		tok = g_scanner_get_next_token(scan);
//		d3_printf("state %d ", state);
//		print_token(scan, tok);
		switch(state) {
		case RCP_STATE_START:
			if (tok == '(')
				state = RCP_STATE_TOPALIST;
			break; /* we just skip stuff up to the first '(' */
		case RCP_STATE_TOPALIST:
			switch(tok) {
			case G_TOKEN_SYMBOL:
				if ((intval(scan) == SYM_OBSERVATION)
					|| (intval(scan) == SYM_RECIPY) ) {
					tok = g_scanner_get_next_token(scan);
					if (tok != '(') 
						break;
					if (io >= no) {
						skip_list(scan);
						obs = NULL;
						break;
					}
					obs = obs_data_new();
					ret = parse_obs(scan, obs);
					if (ret) {
						obs_data_release(obs);
						obs = NULL;
					} else {
						obst[io] = obs;
						io++;
					}
				} else if (intval(scan) == SYM_STARS){
					tok = g_scanner_get_next_token(scan);
					if (tok != '(') 
						break;
					state = RCP_STATE_STARLIST;
				} else {
					skip_one(scan);
				}
				break;
			case ')':
				state = RCP_STATE_START;
				break;
			case '(':
				skip_list(scan);
				break;
			default: /* print a warning here */
				break;
			}
			break;
		case RCP_STATE_STARLIST:
			if (tok == '(') {
				if (i >= n) {
					skip_list(scan);
					break;
				}
				cats = cat_star_new();
				if (cats == NULL) {
					skip_list(scan);
					break;
				}
				if (obs && (obs->flags & OBSDATA_EQUINOX_VALID)) {
					cats->equinox = obs->equinox;
				} else {
					cats->equinox = 2000.0;
				}
				if (!parse_star(scan, cats)) {
					cst[i] = cats;
					i++;
					if (obs) {
						obs_data_ref(obs);
						cats->data = obs;
					}
				} else {
					cat_star_release(cats);
				}
//				d3_printf("end of star\n");
				break;
			} else if (tok == ')') {
				state = RCP_STATE_TOPALIST;
				break;
			} else {
				break;
			}
			break;
		default:
			break;
		}
	} while (tok != G_TOKEN_EOF);
	g_scanner_destroy(scan);
	return i;

}


static int parse_obs(GScanner *scan, struct obs_data *obs)
{
	GTokenType tok;
	int ret;

	double mjd = 0.0;

	do {
		tok = g_scanner_get_next_token(scan);
//		d3_printf("obs ");
//		print_token(scan, tok);
		if (tok == ')')
			continue;
		if (tok != G_TOKEN_SYMBOL) {
			skip_one(scan);
			continue;
		}
		switch(intval(scan)) {
		case SYM_RA:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case G_TOKEN_STRING:
				ret = dms_to_degrees(stringval(scan), &obs->ra);
				if (!ret) {
					obs->ra *= 15.0;
					obs->flags |= OBSDATA_RA_VALID;
				}
				break;
			case G_TOKEN_FLOAT:
				obs->ra = floatval(scan);
				obs->flags |= OBSDATA_RA_VALID;
				break;
			case '(':
				skip_list(scan);
				break;
			case ')':
				continue;
			default:
				break;
			}
			break;
		case SYM_DEC:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case G_TOKEN_STRING:
				ret = dms_to_degrees(stringval(scan), &obs->dec);
				if (!ret) {
					obs->flags |= OBSDATA_DEC_VALID;
				}
				break;
			case '-':
				tok = g_scanner_get_next_token(scan);
				if (tok != G_TOKEN_FLOAT)
					break;
				obs->dec = -floatval(scan);
				obs->flags |= OBSDATA_DEC_VALID;
				break;
			case G_TOKEN_FLOAT:
				obs->dec = floatval(scan);
				obs->flags |= OBSDATA_DEC_VALID;
				break;
			case '(':
				skip_list(scan);
				break;
			case ')':
				continue;
			default:
				break;
			}
			break;
		case SYM_EQUINOX:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case '-':
				tok = g_scanner_get_next_token(scan);
				if (tok != G_TOKEN_FLOAT)
					break;
				obs->equinox = -floatval(scan);
				obs->flags |= OBSDATA_EQUINOX_VALID;
				break;
			case G_TOKEN_FLOAT:
				obs->equinox = floatval(scan);
				obs->flags |= OBSDATA_EQUINOX_VALID;
				break;
			case '(':
				skip_list(scan);
				break;
			case ')':
				continue;
			default:
				break;
			}
			break;
		case SYM_AIRMASS:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case G_TOKEN_FLOAT:
				obs->airmass = floatval(scan);
				obs->flags |= OBSDATA_AIRMASS_VALID;
				break;
			case '(':
				skip_list(scan);
				break;
			case ')':
				continue;
			default:
				break;
			}
			break;
		case SYM_MJD:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case '-':
				tok = g_scanner_get_next_token(scan);
				if (tok != G_TOKEN_FLOAT)
					break;
				mjd = -floatval(scan);
				obs->mjd = mjd;
				obs->flags |= OBSDATA_MJD_VALID;
				break;
			case G_TOKEN_FLOAT:
				mjd = floatval(scan);
				obs->mjd = mjd;
				obs->flags |= OBSDATA_MJD_VALID;
				break;
			case '(':
				skip_list(scan);
				break;
			case ')':
				continue;
			default:
				break;
			}
			break;
		case SYM_FILTER:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case G_TOKEN_STRING:
				obs->filter = strdup(stringval(scan));
				break;
			case '(':
				skip_list(scan);
				break;
			case ')':
				continue;
			default:
				break;
			}
			break;
		case SYM_COMMENTS:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case G_TOKEN_STRING:
				obs->comment = strdup(stringval(scan));
				break;
			case '(':
				skip_list(scan);
				break;
			case ')':
				continue;
			default:
				break;
			}
			break;
		default:
			tok = g_scanner_get_next_token(scan);
			switch(tok) {
			case '(':
				skip_list(scan);
				break;
			default:
				break;
			}
			break;
		}
	} while (tok != ')' && tok != G_TOKEN_EOF);
/* we create an observation name here */
	ret = random();
	if (mjd != 0.0)
		asprintf(&obs->objname, "%c%c%c%.0f", 'a' + ret%26, 'a' + (ret/100)%26, 
			 'a' + (ret/1000)%26, mjd);
	else 
		asprintf(&obs->objname, "%c%c%c", 'a' + ret%26, 'a' + (ret/100)%26, 
			 'a' + (ret/1000)%26);
	return 0;
}
